//
//  Matrix.c
//  BreakBlocks
//
//  Created by Dayle Flowers on 1/5/12.
//  Copyright (c) 2012 Zimmdot. All rights reserved.
//

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <OpenGLES/ES2/gl.h>
#include <OpenGLES/ES2/glext.h>
#include "Matrix.h"

#define O(y,x) (y + (x<<2))

inline void NormalizeVector(float *pVectorX, float *pVectorY, float *pVectorZ)
{
	float normalizingConstant = 1.0 / sqrtf((*pVectorX) * (*pVectorX) + (*pVectorY) * (*pVectorY) + (*pVectorZ) * (*pVectorZ));
	(*pVectorX) *= normalizingConstant;
	(*pVectorY) *= normalizingConstant;
	(*pVectorZ) *= normalizingConstant;
}

inline void ComputeNormalOfPlane(float *pNormalX, float *pNormalY, float *pNormalZ, const float fVector1X, const float fVector1Y, const float fVector1Z, const float fVector2X, const float fVector2Y, const float fVector2Z)
{
	(*pNormalX) = (fVector1Y * fVector2Z) - (fVector1Z * fVector2Y);
	(*pNormalY) = (fVector1Z * fVector2X) - (fVector1X * fVector2Z);
	(*pNormalZ) = (fVector1X * fVector2Y) - (fVector1Y * fVector2X);
}

Matrix4* Matrix4GenPerspective(Matrix4 *pResult, float fFOV, float fAspectRatio, float fNearPlane, float fFarPlane)
{
	float cotangent;
	
	cotangent = 1.0f / tanf(fFOV * M_PI / 360.0);
	
	pResult->m[0] = cotangent / fAspectRatio;
	pResult->m[1] = 0.0;
	pResult->m[2] = 0.0;
	pResult->m[3] = 0.0;
	pResult->m[4] = 0.0;
	pResult->m[5] = cotangent;
	pResult->m[6] = 0.0;
	pResult->m[7] = 0.0;
	pResult->m[8] = 0.0;
	pResult->m[9] = 0.0;
	pResult->m[10] = -(fFarPlane + fNearPlane) / (fFarPlane - fNearPlane);
	pResult->m[11] = -1.0;
	pResult->m[12] = 0.0;
	pResult->m[13] = 0.0;
	pResult->m[14] = -(2 * fFarPlane * fNearPlane) / (fFarPlane - fNearPlane);
	pResult->m[15] = 0.0;
	
	return pResult;
}

Matrix4* Matrix4GenOrtho(Matrix4 *pResult, float fLeft, float fRight, float fTop, float fBottom, float fNearPlane, float fFarPlane)
{
    float deltaX = fRight - fLeft;
    float deltaY = fTop - fBottom;
    float deltaZ = fFarPlane - fNearPlane;
	
    if ( (deltaX == 0.0f) || (deltaY == 0.0f) || (deltaZ == 0.0f) )
        return pResult;

	pResult->m[0] = 2.0f / deltaX;
	pResult->m[1] = 0.0;
	pResult->m[2] = 0.0;
	pResult->m[3] = 0.0;
	pResult->m[4] = 0.0;
	pResult->m[5] = 2.0f / deltaY;
	pResult->m[6] = 0.0;
	pResult->m[7] = 0.0;
	pResult->m[8] = 0.0;
	pResult->m[9] = 0.0;
	pResult->m[10] = -2.0f / deltaZ;
	pResult->m[11] = 0.0;
	pResult->m[12] = -(fRight + fLeft) / deltaX;
	pResult->m[13] = -(fTop + fBottom) / deltaY;
	pResult->m[14] = -(fNearPlane + fFarPlane) / deltaZ;
	pResult->m[15] = 1.0;

	return pResult;
}

Matrix4* Matrix4GenLookAt(Matrix4 *pResult, float eyePositionX, float eyePositionY, float eyePositionZ, float centerX, float centerY, float centerZ, float upVectorX, float upVectorY, float upVectorZ)
{
	// Compute Forward
	float forwardX = centerX - eyePositionX;
	float forwardY = centerY - eyePositionY;
	float forwardZ = centerZ - eyePositionZ;
	NormalizeVector(&forwardX, &forwardY, &forwardZ);

	// Side = Forward x Up
	ComputeNormalOfPlane(&pResult->m[0], &pResult->m[4], &pResult->m[8], forwardX, forwardY, forwardZ, upVectorX, upVectorY, upVectorZ);
	NormalizeVector(&pResult->m[0], &pResult->m[4], &pResult->m[8]);

	// Recompute up as: Up = Side x Forward
	ComputeNormalOfPlane(&pResult->m[1], &pResult->m[5], &pResult->m[9], pResult->m[0], pResult->m[4], pResult->m[8], forwardX, forwardY, forwardZ);

	pResult->m[2] = -forwardX;
	pResult->m[3] = 0.0;
	pResult->m[6] = -forwardY;
	pResult->m[7] = 0.0;
	pResult->m[10] = -forwardZ;
	pResult->m[11] = 0.0;
	pResult->m[12] = 0.0;
	pResult->m[13] = 0.0;
	pResult->m[14] = 0.0;
	pResult->m[15] = 1.0;

	Matrix4Translate(pResult, -eyePositionX, -eyePositionY, -eyePositionZ);
	
	return pResult;
}

Matrix4* Matrix4LoadIdentity(Matrix4 *pResult)
{
	pResult->m[0] = 1.0;
	pResult->m[1] = 0.0;
	pResult->m[2] = 0.0;
	pResult->m[3] = 0.0;
	pResult->m[4] = 0.0;
	pResult->m[5] = 1.0;
	pResult->m[6] = 0.0;
	pResult->m[7] = 0.0;
	pResult->m[8] = 0.0;
	pResult->m[9] = 0.0;
	pResult->m[10] = 1.0;
	pResult->m[11] = 0.0;
	pResult->m[12] = 0.0;
	pResult->m[13] = 0.0;
	pResult->m[14] = 0.0;
	pResult->m[15] = 1.0;
	
	return pResult;	
}

Matrix4* Matrix4MakeFromArray(Matrix4 *pResult, float *pArray)
{
	pResult->m[0] = pArray[0];
	pResult->m[1] = pArray[1];
	pResult->m[2] = pArray[2];
	pResult->m[3] = pArray[3];
	pResult->m[4] = pArray[4];
	pResult->m[5] = pArray[5];
	pResult->m[6] = pArray[6];
	pResult->m[7] = pArray[7];
	pResult->m[8] = pArray[8];
	pResult->m[9] = pArray[9];
	pResult->m[10] = pArray[10];
	pResult->m[11] = pArray[11];
	pResult->m[12] = pArray[12];
	pResult->m[13] = pArray[13];
	pResult->m[14] = pArray[14];
	pResult->m[15] = pArray[15];
	
	return pResult;
}

Matrix4* Matrix4MakeTranslation(Matrix4 *pResult, float fX, float fY, float fZ)
{
	pResult->m[0] = 1.0;
	pResult->m[1] = 0.0;
	pResult->m[2] = 0.0;
	pResult->m[3] = 0.0;
	pResult->m[4] = 0.0;
	pResult->m[5] = 1.0;
	pResult->m[6] = 0.0;
	pResult->m[7] = 0.0;
	pResult->m[8] = 0.0;
	pResult->m[9] = 0.0;
	pResult->m[10] = 1.0;
	pResult->m[11] = 0.0;
	pResult->m[12] = fX;
	pResult->m[13] = fY;
	pResult->m[14] = fZ;
	pResult->m[15] = 1.0;
	
	return pResult;
}

Matrix4* Matrix4MakeRotation(Matrix4 *pResult, float fRotation, float fX, float fY, float fZ)
{
	float c, s, c1, mag;
	
	mag = sqrtf(fX * fX + fY * fY + fZ * fZ);
	
	if (mag > 0.0f)
	{
		fX /= mag;
		fY /= mag;
		fZ /= mag;
		
		fRotation *= M_PI / 360.0f;
		
		c = cosf(fRotation);
		s = sinf(fRotation);
		c1 = 1.0f - c;

		pResult->m[0] = fX * fX * c1 + c;
		pResult->m[1] = fX * fY * c1 + fZ * s;
		pResult->m[2] = fX * fZ * c1 - fY * s;
		pResult->m[3] = 0.0;
		pResult->m[4] = fX * fY * c1 - fZ * s;
		pResult->m[5] = fY * fY * c1 + c;
		pResult->m[6] = fY * fZ * c1 + fX * s;
		pResult->m[7] = 0.0;
		pResult->m[8] = fX * fZ * c1 + fY * s;
		pResult->m[9] = fY * fZ * c1 - fX * s;
		pResult->m[10] = fZ * fZ * c1 + c;
		pResult->m[11] = 0.0;
		pResult->m[12] = 0.0;
		pResult->m[13] = 0.0;
		pResult->m[14] = 0.0;
		pResult->m[15] = 1.0;
	}

	return pResult;
}

Matrix4* Matrix4Translate(Matrix4 *pInOutResult, float fX, float fY, float fZ)
{
    pInOutResult->m[12] += (pInOutResult->m[0] * fX + pInOutResult->m[4] * fY + pInOutResult->m[8] * fZ);
    pInOutResult->m[13] += (pInOutResult->m[1] * fX + pInOutResult->m[5] * fY + pInOutResult->m[9] * fZ);
    pInOutResult->m[14] += (pInOutResult->m[2] * fX + pInOutResult->m[6] * fY + pInOutResult->m[10] * fZ);
    pInOutResult->m[15] += (pInOutResult->m[3] * fX + pInOutResult->m[7] * fY + pInOutResult->m[11] * fZ);
	
	return pInOutResult;
}

Matrix4* Matrix4Rotate(Matrix4 *pInOutResult, float fRotation, float fX, float fY, float fZ)
{
	Matrix4 temp, temp2;
	
	memcpy(temp.m, pInOutResult->m, 16 * sizeof(float));
	
	return Matrix4Multiply(pInOutResult, &temp, Matrix4MakeRotation(&temp2, fRotation, fX, fY, fZ));
}

Matrix4* Matrix4Multiply(Matrix4 *__restrict pResult, Matrix4 *__restrict pMat1, Matrix4 *__restrict pMat2)
{
	float *src1, *src2, *dest;
	
	src1 = pMat1->m;
	src2 = pMat2->m;
	dest = pResult->m;
	
    *(dest+O(0,0)) = (*(src1+O(0,0)) * *(src2+O(0,0))) + (*(src1+O(0,1)) * *(src2+O(1,0))) + (*(src1+O(0,2)) * *(src2+O(2,0))) + (*(src1+O(0,3)) * *(src2+O(3,0)));
    *(dest+O(0,1)) = (*(src1+O(0,0)) * *(src2+O(0,1))) + (*(src1+O(0,1)) * *(src2+O(1,1))) + (*(src1+O(0,2)) * *(src2+O(2,1))) + (*(src1+O(0,3)) * *(src2+O(3,1)));
    *(dest+O(0,2)) = (*(src1+O(0,0)) * *(src2+O(0,2))) + (*(src1+O(0,1)) * *(src2+O(1,2))) + (*(src1+O(0,2)) * *(src2+O(2,2))) + (*(src1+O(0,3)) * *(src2+O(3,2)));
    *(dest+O(0,3)) = (*(src1+O(0,0)) * *(src2+O(0,3))) + (*(src1+O(0,1)) * *(src2+O(1,3))) + (*(src1+O(0,2)) * *(src2+O(2,3))) + (*(src1+O(0,3)) * *(src2+O(3,3)));
    *(dest+O(1,0)) = (*(src1+O(1,0)) * *(src2+O(0,0))) + (*(src1+O(1,1)) * *(src2+O(1,0))) + (*(src1+O(1,2)) * *(src2+O(2,0))) + (*(src1+O(1,3)) * *(src2+O(3,0)));
    *(dest+O(1,1)) = (*(src1+O(1,0)) * *(src2+O(0,1))) + (*(src1+O(1,1)) * *(src2+O(1,1))) + (*(src1+O(1,2)) * *(src2+O(2,1))) + (*(src1+O(1,3)) * *(src2+O(3,1)));
    *(dest+O(1,2)) = (*(src1+O(1,0)) * *(src2+O(0,2))) + (*(src1+O(1,1)) * *(src2+O(1,2))) + (*(src1+O(1,2)) * *(src2+O(2,2))) + (*(src1+O(1,3)) * *(src2+O(3,2)));
    *(dest+O(1,3)) = (*(src1+O(1,0)) * *(src2+O(0,3))) + (*(src1+O(1,1)) * *(src2+O(1,3))) + (*(src1+O(1,2)) * *(src2+O(2,3))) + (*(src1+O(1,3)) * *(src2+O(3,3)));
    *(dest+O(2,0)) = (*(src1+O(2,0)) * *(src2+O(0,0))) + (*(src1+O(2,1)) * *(src2+O(1,0))) + (*(src1+O(2,2)) * *(src2+O(2,0))) + (*(src1+O(2,3)) * *(src2+O(3,0)));
    *(dest+O(2,1)) = (*(src1+O(2,0)) * *(src2+O(0,1))) + (*(src1+O(2,1)) * *(src2+O(1,1))) + (*(src1+O(2,2)) * *(src2+O(2,1))) + (*(src1+O(2,3)) * *(src2+O(3,1)));
    *(dest+O(2,2)) = (*(src1+O(2,0)) * *(src2+O(0,2))) + (*(src1+O(2,1)) * *(src2+O(1,2))) + (*(src1+O(2,2)) * *(src2+O(2,2))) + (*(src1+O(2,3)) * *(src2+O(3,2)));
    *(dest+O(2,3)) = (*(src1+O(2,0)) * *(src2+O(0,3))) + (*(src1+O(2,1)) * *(src2+O(1,3))) + (*(src1+O(2,2)) * *(src2+O(2,3))) + (*(src1+O(2,3)) * *(src2+O(3,3)));
    *(dest+O(3,0)) = (*(src1+O(3,0)) * *(src2+O(0,0))) + (*(src1+O(3,1)) * *(src2+O(1,0))) + (*(src1+O(3,2)) * *(src2+O(2,0))) + (*(src1+O(3,3)) * *(src2+O(3,0)));
    *(dest+O(3,1)) = (*(src1+O(3,0)) * *(src2+O(0,1))) + (*(src1+O(3,1)) * *(src2+O(1,1))) + (*(src1+O(3,2)) * *(src2+O(2,1))) + (*(src1+O(3,3)) * *(src2+O(3,1)));
    *(dest+O(3,2)) = (*(src1+O(3,0)) * *(src2+O(0,2))) + (*(src1+O(3,1)) * *(src2+O(1,2))) + (*(src1+O(3,2)) * *(src2+O(2,2))) + (*(src1+O(3,3)) * *(src2+O(3,2)));
    *(dest+O(3,3)) = (*(src1+O(3,0)) * *(src2+O(0,3))) + (*(src1+O(3,1)) * *(src2+O(1,3))) + (*(src1+O(3,2)) * *(src2+O(2,3))) + (*(src1+O(3,3)) * *(src2+O(3,3)));
	
	return pResult;
}

Matrix3* Matrix4ToMatrix3(Matrix3 *pResult, Matrix4 *pSource)
{
	pResult->m[0] = pSource->m[0];
	pResult->m[1] = pSource->m[1];
	pResult->m[2] = pSource->m[2];
	pResult->m[3] = pSource->m[4];
	pResult->m[4] = pSource->m[5];
	pResult->m[5] = pSource->m[6];
	pResult->m[6] = pSource->m[8];
	pResult->m[7] = pSource->m[9];
	pResult->m[8] = pSource->m[10];

	return pResult;
}

Matrix3* Matrix3Invert(Matrix3 *__restrict pResult, Matrix3 *__restrict pSource)
{
	float det;
	
	det = 1.0f / (pSource->m[0] * pSource->m[4] * pSource->m[8] -
				  pSource->m[0] * pSource->m[7] * pSource->m[5] -
				  pSource->m[3] * pSource->m[1] * pSource->m[8] +
				  pSource->m[3] * pSource->m[7] * pSource->m[2] +
				  pSource->m[6] * pSource->m[1] * pSource->m[5] -
				  pSource->m[6] * pSource->m[4] * pSource->m[2]);
	
	pResult->m[0] = det * (pSource->m[4] * pSource->m[8] - pSource->m[7] * pSource->m[5]);
	pResult->m[1] = det * (pSource->m[7] * pSource->m[2] - pSource->m[1] * pSource->m[8]);
	pResult->m[2] = det * (pSource->m[1] * pSource->m[5] - pSource->m[4] * pSource->m[2]);
	pResult->m[3] = det * (pSource->m[6] * pSource->m[5] - pSource->m[3] * pSource->m[8]);
	pResult->m[4] = det * (pSource->m[0] * pSource->m[8] - pSource->m[6] * pSource->m[2]);
	pResult->m[5] = det * (pSource->m[3] * pSource->m[2] - pSource->m[0] * pSource->m[5]);
	pResult->m[6] = det * (pSource->m[3] * pSource->m[7] - pSource->m[6] * pSource->m[4]);
	pResult->m[7] = det * (pSource->m[6] * pSource->m[1] - pSource->m[0] * pSource->m[7]);
	pResult->m[8] = det * (pSource->m[0] * pSource->m[4] - pSource->m[3] * pSource->m[1]);	
	
	return pResult;
}

Matrix3* Matrix3InvertAndTranspose(Matrix3 *__restrict pResult, Matrix3 *__restrict pSource)
{
	Matrix3 temp;
	
	// Invert
	Matrix3Invert(&temp, pSource);
	
	// Transpose
	pResult->m[0] = temp.m[0];
	pResult->m[1] = temp.m[3];
	pResult->m[2] = temp.m[6];
	pResult->m[3] = temp.m[1];
	pResult->m[4] = temp.m[4];
	pResult->m[5] = temp.m[7];
	pResult->m[6] = temp.m[2];
	pResult->m[7] = temp.m[5];
	pResult->m[8] = temp.m[8];
	
	return pResult;
}

