/////////////////////////////////////////////////////////////////
// Name: xGraphics.h
/////////////////////////////////////////////////////////////////
// Description:
// - Core graphics system that handles rendering of all
//   xRenderable objects in the current xSceneGraph object
/////////////////////////////////////////////////////////////////
// - 1.8.2004, Update, Arto Ruotsalainen (Arex)
//   Added iFrame to CreateTexture, for single textures
//   just set it to zero, it is for multiple frame objects. 
//
// - 15.7.2004, Update, Arto Ruotsalainen (Arex)
//   Added get device pointer function
//
// - 13.7.2004, Update, Arto Ruotsalainen (Arex)
//   Added screen width and height variables.
//
// - 7.10.2004, Update, Dayle Flowers(Xienen)
//   Added xPool for Objects and Create2DObject() function
//   Added AlphaBlit and Goto2DMode functions
//
// - 7.6.2004, Update, Arto Ruotsalainen(Arex)
//   Comment stuff added to begin of the file.
//   Add new updates/fixes above this, so the old updates
//   will be at the bottom.
/////////////////////////////////////////////////////////////////
#ifndef XGRAPHICS_H_
#define XGRAPHICS_H_

#include <stdlib.h>
#include <OpenGLES/ES2/gl.h>
#include <OpenGLES/ES2/glext.h>
#include <OpenGLES/EAGL.h>

#include "xResults.h"
#include "xTexture.h"
#include "xVector.h"

#include "xRenderData.h"
#include "xSceneGraph.h"

#define BB_GL_QUADS	7

typedef UInt32 DWORD;

//#pragma comment(lib, "opengl32.lib")
//#pragma comment(lib, "glu32.lib")
//#pragma comment(lib, "glaux.lib")

struct xGraphics
{
	void*		m_pWindow;

	int			m_iScreenWidth;
	int			m_iScreenHeight;
	Matrix4		m_matProjection;
	Matrix4		m_matPerspective;
	DWORD		m_dwDisplayFrequency;
	float		m_fGrayScaleFactor;
	CFArrayRef	m_arrayDisplayModes;
	int			m_iDisplayModeCount;
	int			m_iAllScreenResolutionsIndex;

	EAGLContext	*m_hContext;
	xSceneGraph	*m_pSceneGraph;
	xTexture	 m_xTexturePool;
	
	GLuint	m_uiRectIndexBuffer;
	GLuint	m_ui3RectIndexBuffer;
	GLuint	m_ui4RectIndexBuffer;
	GLuint	m_ui9RectIndexBuffer;

	int m_iFrameSkip;

	GLuint m_uiAlphaKillPixelProgramIndex;
	GLuint m_uiAlphaMaskPixelProgramIndex;
	GLuint m_uiGrayScalePixelProgramIndex;
	GLuint m_uiDualColorPixelProgramIndex;

	// Alpha Kill Pixel shader
	struct
	{
		GLint m_iTexture;
	} m_xAlphaKillPixel;

	// Alpha Mask Pixel shader
	struct
	{
		GLint m_iTextureRatio;
		GLint m_iTextureOffset;
		GLint m_iGrayScaleFactor;
		GLint m_iMask;
		GLint m_iTexture;
	} m_xAlphaMaskPixel;

	// GrayScale Pixel shader
	struct
	{
		GLint m_iTexture;
		GLint m_iPosition;
		GLint m_iTextureCoord;
		GLint m_iMVP;
		GLint m_iColor;
	} m_xGrayScalePixel;

	// Dual Color Pixel shader
	struct
	{
		GLint m_iTexture;
		GLint m_iPosition;
		GLint m_iTextureCoord;
		GLint m_iMVP;
		GLint m_iSecondColor;
		GLint m_iColorDifference;
	} m_xDualColorPixel;
};

extern xGraphics g_xGraphics;

FORCE_INLINE void xGraphicsSetSceneGraph(xSceneGraph *pSceneGraph) { g_xGraphics.m_pSceneGraph = pSceneGraph; }

int xGraphicsInit(void* pWindow, int iWindowWidth, int iWindowHeight, bool bFullscreenApp, int iBPP);
void xGraphicsUpdate(double dTime);
void xGraphicsShutdown();

void xGraphicsGetDesktopResolution(int *pScreenWidth, int *pScreenHeight);
void xGraphicsStartGettingAllDesktopResolutions();
bool xGraphicsGetNextDesktopResolution(int *pScreenWidth, int *pScreenHeight);
bool xGraphicsIsResolutionAvailable(int iScreenWidth, int iScreenHeight, bool bFullscreen);

xTexture* xGraphicsGetTexture(const char* strFilePath, int iFrame);
int xGraphicsGetTextures(const char* strFilePath, xTexture ***pTextures, float *fAnimationRate = 0);
void xGraphicsReloadTexturePool();
void xGraphicsDeleteTexture(xTexture* pTexture);
void xGraphicsDeleteTextures(xTexture **pTextures, int iCount);

FORCE_INLINE void xGraphicsDeleteTextureReference(xTexture* pTexture)
{
	//DEBUGLOG("xGraphicsDeleteTextureRef: %s - %i\n", pTexture->m_strObjectPath, pTexture->m_iReferenceCount);

	if(pTexture->m_iReferenceCount > 0)
		pTexture->m_iReferenceCount--;
}

FORCE_INLINE xTexture* xGraphicsTexturePoolNew(const char* strFilePath)
{
	xTexture *current = &g_xGraphics.m_xTexturePool;

	while(current->m_pNext)
	{
		if(!stricmp(current->m_strObjectPath, strFilePath))
		{
			current->m_iReferenceCount++;
			//DEBUGLOG("xGraphicsTexturePoolNew: %s - %i\n", current->m_strObjectPath, current->m_iReferenceCount);
			return current;
		}

		current = current->m_pNext;
	}

	if(!stricmp(current->m_strObjectPath, strFilePath))
	{
		current->m_iReferenceCount++;
		//DEBUGLOG("xGraphicsTexturePoolNew: %s - %i\n", current->m_strObjectPath, current->m_iReferenceCount);
		return current;
	}

	current->m_pNext = (xTexture*)dlmalloc(sizeof(xTexture));
	xTextureInit(current->m_pNext);
	current->m_pNext->m_pPrev = current;
	
	strcpy(current->m_pNext->m_strObjectPath, strFilePath);

	//DEBUGLOG("xGraphicsTexturePoolNew: %s - %i\n", current->m_pNext->m_strObjectPath, current->m_pNext->m_iReferenceCount);

	return current->m_pNext;
}

FORCE_INLINE xTexture* xGraphicsTexturePoolGet(const char* strFilePath)
{
	xTexture *current = &g_xGraphics.m_xTexturePool;

	while(current->m_pNext)
	{
		if(!stricmp(current->m_strObjectPath, strFilePath))
		{
			current->m_iReferenceCount++;
			//DEBUGLOG("xGraphicsTexturePoolGet: %s - %i\n", current->m_strObjectPath, current->m_iReferenceCount);
			return current;
		}

		current = current->m_pNext;
	}

	if(!stricmp(current->m_strObjectPath, strFilePath))
	{
		current->m_iReferenceCount++;
		//DEBUGLOG("xGraphicsTexturePoolGet: %s - %i\n", current->m_strObjectPath, current->m_iReferenceCount);
		return current;
	}

	return 0;
}

#endif